/* (C) 2009-2011 Yana */

#if !defined(YK0A_INTERNAL_H)
#define YK0A_INTERNAL_H

#include "yk0a_base.h"

#define YK0A_CONTEXT_BASE 8
#if defined(YK0A_CUSTOM_ENABLE_PORT)
#define YK0A_CONTEXT_SIZE 19
#else
#define YK0A_CONTEXT_SIZE 18
#endif

#define YK0A_STATE_READY 0
#define YK0A_STATE_RECEIVE 1
#define YK0A_STATE_REPLY 2
#define YK0A_STATE_SEND 3

#define YK0A_ARM_CPSR_I (1 << 7)
#define YK0A_ARM_CPSR_M_MASK (0x1f << YK0A_ARM_CPSR_M_SHIFT)
#define YK0A_ARM_CPSR_M_SHIFT 0

#define YK0A_ARM_CPSR_M_IRQ (0x12 << YK0A_ARM_CPSR_M_SHIFT)
#define YK0A_ARM_CPSR_M_SVC (0x13 << YK0A_ARM_CPSR_M_SHIFT)
#define YK0A_ARM_CPSR_M_SYS (0x1f << YK0A_ARM_CPSR_M_SHIFT)

#define YK0A_INTERRUPT_ADDRESS_STATUS 0x16000000
#define YK0A_INTERRUPT_ADDRESS_MASK 0x16000004
#define YK0A_INTERRUPT_ADDRESS_UNMASK 0x16000008
#define YK0A_INTERRUPT_SIZE 32

#if !defined(YK0A_PREPROCESS_ONLY)

#if defined(YK0A_CUSTOM_DISABLE_ASM)

yk0a_uint32_t yk0a_arm_read_cpsr(void);

#else

static yk0a_uint32_t yk0a_arm_read_cpsr(void) {
	yk0a_uint32_t cpsr;
	__asm__ __volatile__("mrs %0, cpsr" : "=r"(cpsr));
	return cpsr;
}

#endif

yk0a_uint32_t yk0a_interrupt_mask;

static yk0a_uint32_t yk0a_interrupt_read_mask(void) {
	return yk0a_interrupt_mask;
}

static yk0a_uint32_t yk0a_interrupt_read_request(void) {
	return *(volatile yk0a_uint32_t *)YK0A_INTERRUPT_ADDRESS_STATUS;
}

static void yk0a_interrupt_write_mask(yk0a_uint32_t mask) {
	yk0a_uint32_t i;
	for (i = 0; i < YK0A_INTERRUPT_SIZE; ++i) {
		if ((1 << i & mask) != (1 << i & yk0a_interrupt_mask)) {
			if ((1 << i & mask) == 0) {
				*(volatile yk0a_uint32_t *)YK0A_INTERRUPT_ADDRESS_MASK = i;
			} else {
				*(volatile yk0a_uint32_t *)YK0A_INTERRUPT_ADDRESS_UNMASK = i;
			}
		}
	}
	yk0a_interrupt_mask = mask;
}

static void yk0a_interrupt_write_request(yk0a_uint32_t request) {
}

static void yk0a_interrupt_initialize(void) {
	yk0a_interrupt_mask = ~(yk0a_uint32_t)0;
	yk0a_interrupt_write_mask(0);
}

#endif

#endif
