/* (C) 2009-2011 Yana */

#if !defined(YK0A_INTERNAL_H)
#define YK0A_INTERNAL_H

#include "yk0a_base.h"

#define YK0A_CONTEXT_BASE 8
#if defined(YK0A_CUSTOM_ENABLE_PORT)
#define YK0A_CONTEXT_SIZE 35
#else
#define YK0A_CONTEXT_SIZE 34
#endif

#define YK0A_STATE_READY 0
#define YK0A_STATE_RECEIVE 1
#define YK0A_STATE_REPLY 2
#define YK0A_STATE_SEND 3

#define YK0A_MIPS_STATUS $12
#define YK0A_MIPS_CAUSE $13
#define YK0A_MIPS_EPC $14

#define YK0A_MIPS_HAZARD_STATUS ssnop; ssnop
#define YK0A_MIPS_HAZARD_CAUSE ssnop
#define YK0A_MIPS_HAZARD_EPC

#define YK0A_MIPS_STATUS_IM_MASK (0xff << YK0A_MIPS_STATUS_IM_SHIFT)
#define YK0A_MIPS_STATUS_IM_SHIFT 8
#define YK0A_MIPS_STATUS_KSU_MASK (3 << YK0A_MIPS_STATUS_KSU_SHIFT)
#define YK0A_MIPS_STATUS_KSU_SHIFT 3
#define YK0A_MIPS_STATUS_ERL (1 << 2)
#define YK0A_MIPS_STATUS_EXL (1 << 1)
#define YK0A_MIPS_STATUS_IE (1 << 0)

#define YK0A_MIPS_CAUSE_BD (1 << 31)
#define YK0A_MIPS_CAUSE_IP_MASK (0xff << YK0A_MIPS_CAUSE_IP_SHIFT)
#define YK0A_MIPS_CAUSE_IP_SHIFT 8
#define YK0A_MIPS_CAUSE_EXCCODE_MASK (0x1f << YK0A_MIPS_CAUSE_EXCCODE_SHIFT)
#define YK0A_MIPS_CAUSE_EXCCODE_SHIFT 2

#define YK0A_MIPS_CAUSE_EXCCODE_INT (0x00 << YK0A_MIPS_CAUSE_EXCCODE_SHIFT)
#define YK0A_MIPS_CAUSE_EXCCODE_SYS (0x08 << YK0A_MIPS_CAUSE_EXCCODE_SHIFT)

#if !defined(YK0A_PREPROCESS_ONLY)

#if defined(YK0A_CUSTOM_DISABLE_ASM)

yk0a_uint32_t yk0a_mips_read_cause(void);
yk0a_uint32_t yk0a_mips_read_status(void);
void yk0a_mips_write_cause(yk0a_uint32_t cause);
void yk0a_mips_write_status(yk0a_uint32_t status);

#else

static yk0a_uint32_t yk0a_mips_read_cause(void) {
	yk0a_uint32_t cause;
	__asm__ __volatile__("mfc0 %0, " YK0A_PREPROCESS_STRING(YK0A_MIPS_CAUSE) : "=r"(cause));
	return cause;
}

static yk0a_uint32_t yk0a_mips_read_status(void) {
	yk0a_uint32_t status;
	__asm__ __volatile__("mfc0 %0, " YK0A_PREPROCESS_STRING(YK0A_MIPS_STATUS) : "=r"(status));
	return status;
}

static void yk0a_mips_write_cause(yk0a_uint32_t cause) {
	__asm__ __volatile__("mtc0 %0, " YK0A_PREPROCESS_STRING(YK0A_MIPS_CAUSE) "; " YK0A_PREPROCESS_STRING(YK0A_MIPS_HAZARD_CAUSE) : : "r"(cause));
}

static void yk0a_mips_write_status(yk0a_uint32_t status) {
	__asm__ __volatile__("mtc0 %0, " YK0A_PREPROCESS_STRING(YK0A_MIPS_STATUS) "; " YK0A_PREPROCESS_STRING(YK0A_MIPS_HAZARD_STATUS) : : "r"(status));
}

#endif

static yk0a_uint32_t yk0a_interrupt_read_mask(void) {
	return (yk0a_mips_read_status() & YK0A_MIPS_STATUS_IM_MASK) >> YK0A_MIPS_STATUS_IM_SHIFT;
}

static yk0a_uint32_t yk0a_interrupt_read_request(void) {
	return (yk0a_mips_read_cause() & YK0A_MIPS_CAUSE_IP_MASK) >> YK0A_MIPS_CAUSE_IP_SHIFT;
}

static void yk0a_interrupt_write_mask(yk0a_uint32_t mask) {
	yk0a_mips_write_status((yk0a_mips_read_status() & ~YK0A_MIPS_STATUS_IM_MASK) | (mask << YK0A_MIPS_STATUS_IM_SHIFT & YK0A_MIPS_STATUS_IM_MASK));
}

static void yk0a_interrupt_write_request(yk0a_uint32_t request) {
	yk0a_mips_write_cause((yk0a_mips_read_cause() & ~YK0A_MIPS_CAUSE_IP_MASK) | (request << YK0A_MIPS_CAUSE_IP_SHIFT & YK0A_MIPS_CAUSE_IP_MASK));
}

static void yk0a_interrupt_initialize(void) {
	yk0a_interrupt_write_mask(0);
}

#endif

#endif
