/* (C) 2009-2011 Yana */

#if !defined(YK0A_INTERRUPT_H)
#define YK0A_INTERRUPT_H

#include "yk0a_base.h"

#define YK0A_FLAG_INTERRUPT_MASK (0xff << YK0A_FLAG_INTERRUPT_SHIFT)
#define YK0A_FLAG_INTERRUPT_SHIFT 0

#define YK0A_PROCEDURE_INTERRUPT_BIT_MASK 11
#define YK0A_PROCEDURE_INTERRUPT_BIT_REQUEST 12

#if !defined(YK0A_PREPROCESS_ONLY)

#include <stddef.h>

typedef yk0a_uint32_t yk0a_interrupt_t;

#if defined(YK0A_CUSTOM_DISABLE_ASM)

yk0a_error_t yk0a_interrupt_bit_mask(yk0a_interrupt_t mask_off, yk0a_interrupt_t mask_on, yk0a_interrupt_t *mask_out);
yk0a_error_t yk0a_interrupt_bit_request(yk0a_interrupt_t request_off, yk0a_interrupt_t request_on, yk0a_interrupt_t *request_out);

#else

__inline__ static yk0a_error_t yk0a_interrupt_bit_mask(yk0a_interrupt_t mask_off, yk0a_interrupt_t mask_on, yk0a_interrupt_t *mask_out) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_INTERRUPT_BIT_MASK;
	register yk0a_uint32_t gpr3 __asm__("$3") = mask_off;
	register yk0a_uint32_t gpr4 __asm__("$4") = mask_on;
	__asm__ __volatile__("syscall" : "=r"(gpr2), "=r"(gpr3) : "0"(gpr2), "1"(gpr3), "r"(gpr4) : "memory");
	if (mask_out != NULL) {
		*mask_out = gpr3;
	}
	return gpr2;
}

__inline__ static yk0a_error_t yk0a_interrupt_bit_request(yk0a_interrupt_t request_off, yk0a_interrupt_t request_on, yk0a_interrupt_t *request_out) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_INTERRUPT_BIT_REQUEST;
	register yk0a_uint32_t gpr3 __asm__("$3") = request_off;
	register yk0a_uint32_t gpr4 __asm__("$4") = request_on;
	__asm__ __volatile__("syscall" : "=r"(gpr2), "=r"(gpr3) : "0"(gpr2), "1"(gpr3), "r"(gpr4) : "memory");
	if (request_out != NULL) {
		*request_out = gpr3;
	}
	return gpr2;
}

#endif

#endif

#endif
