/* (C) 2009-2011 Yana */

#include "device.h"
#include "start_setup.h"
#include "y_arm.h"
#include "y_base.h"
#include "y_device.h"
#include "y_queue.h"
#include "y_yk0a.h"
#include "yk0a.h"
#include "yk0a_base.h"
#include "yk0a_interrupt.h"
#include "yk0a_setup.h"
#include <stddef.h>

typedef struct {
	y_queue_t queue;
} wrap_task_control_t;

#define WRAP0_CONTEXT_ARGUMENT0 YK0A_CONTEXT_ARGUMENT0
#define WRAP0_CONTEXT_ARGUMENT1 YK0A_CONTEXT_ARGUMENT1

#define WRAP0_PROCEDURE_CONS_GETCHAR 0
#define WRAP0_PROCEDURE_CONS_HALT 1
#define WRAP0_PROCEDURE_CONS_PUTCHAR 2

__inline__ static yk0a_error_t wrap0_cons_getchar(y_uint8_t *c) {
	register yk0a_uint32_t r0 __asm__("r0") = WRAP0_PROCEDURE_CONS_GETCHAR;
	register yk0a_uint32_t r1 __asm__("r1");
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0) : "memory");
	if (c != NULL) {
		*c = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t wrap0_cons_halt(void) {
	register yk0a_uint32_t r0 __asm__("r0") = WRAP0_PROCEDURE_CONS_HALT;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0) : "memory");
	return r0;
}

__inline__ static yk0a_error_t wrap0_cons_putchar(y_uint8_t c) {
	register yk0a_uint32_t r0 __asm__("r0") = WRAP0_PROCEDURE_CONS_PUTCHAR;
	register yk0a_uint32_t r1 __asm__("r1") = c;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1) : "memory");
	return r0;
}

wrap_task_control_t wrap_task_control[YK0A_SETUP_TASK_SIZE];

y_uint8_t task0_stack[TASK0_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));
y_uint8_t task1_stack[TASK1_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));
y_uint8_t task2_stack[TASK2_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));

void task1(void);
void task2(void);

void task0(void) {
	yk0a_flag_t f;
	yk0a_task_t t;
	y_device_cons_initialize();
	y_yk0a_start(1, task1, task1_stack + TASK1_STACK_SIZE / Y_STACK_ALIGNMENT * Y_STACK_ALIGNMENT - Y_STACK_ARGUMENT, y_arm_read_cpsr());
	yk0a_set_context(2, YK0A_CUSTOM_CONTEXT_PORT, YK0A_SETUP_INTERRUPT_PORT, NULL);
	yk0a_set_priority(2, 1, NULL);
	y_yk0a_start(2, task2, task2_stack + TASK2_STACK_SIZE / Y_STACK_ALIGNMENT * Y_STACK_ALIGNMENT - Y_STACK_ARGUMENT, y_arm_read_cpsr());
	yk0a_listen(YK0A_SETUP_EXCEPTION_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
	y_device_cons_put_string("{");
	y_device_cons_put_hexadecimal(f);
	if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_EXCEPTION_PORT, &t) == 0) {
		yk0a_context_value_t cpsr;
		yk0a_get_context(t, YK0A_CONTEXT_CPSR, &cpsr);
		y_device_cons_put_string(" ");
		y_device_cons_put_hexadecimal(t);
		y_device_cons_put_string(" ");
		y_device_cons_put_hexadecimal(cpsr);
	}
	y_device_cons_put_string("}\n");
	y_device_cons_halt();
}

void task1(void) {
	y_uint8_t c;
	y_queue_t q;
	c = 0;
	y_queue_initialize(&q);
	yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		yk0a_listen(YK0A_SETUP_INTERRUPT_PORT, YK0A_FLAG_KERNEL_SENDER | DEVICE_INTERRUPT_CONS, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_INTERRUPT_PORT, &t) == 0) {
			yk0a_context_value_t a0;
			yk0a_context_value_t a1;
			yk0a_set_context(t, WRAP0_CONTEXT_ARGUMENT0, 0, &a0);
			switch (a0) {
			case WRAP0_PROCEDURE_CONS_GETCHAR:
				y_queue_append(&q, (y_queue_t *)&wrap_task_control[t]);
				break;
			case WRAP0_PROCEDURE_CONS_PUTCHAR:
				yk0a_get_context(t, WRAP0_CONTEXT_ARGUMENT1, &a1);
				yk0a_reply(t);
				y_device_cons_putchar(a1);
				break;
			default:
				yk0a_forward(YK0A_SETUP_EXCEPTION_PORT, t);
				break;
			}
		}
		if ((f & DEVICE_INTERRUPT_CONS) != 0) {
			c = y_device_cons_getchar();
			yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
		}
		if (c != 0 && y_queue_copy(&q) != NULL) {
			t = (wrap_task_control_t *)y_queue_remove(&q) - &wrap_task_control[0];
			yk0a_set_context(t, WRAP0_CONTEXT_ARGUMENT1, c, NULL);
			yk0a_reply(t);
			c = 0;
		}
	}
}

void task2(void) {
	while (1) {
		y_uint8_t c;
		wrap0_cons_getchar(&c);
		if (c < ' ') {
			break;
		}
		wrap0_cons_putchar(c);
	}
	wrap0_cons_halt();
}
