/* (C) 2009-2011 Yana */

#include "device.h"
#include "start_setup.h"
#include "y_base.h"
#include "y_device.h"
#include "y_mips.h"
#include "y_queue.h"
#include "y_yk0a.h"
#include "yk0a.h"
#include "yk0a_base.h"
#include "yk0a_interrupt.h"
#include "yk0a_setup.h"
#include <stddef.h>

#define WRAP_PORT_SIZE 1

typedef y_uint32_t wrap_port_t;

typedef struct {
	y_queue_t queue;
	yk0a_flag_t flag;
	yk0a_task_t receiver;
} wrap_port_control_t;

typedef struct {
	y_queue_t queue;
} wrap_task_control_t;

#define WRAP0_CONTEXT_ARGUMENT0 YK0A_CONTEXT_GPR4
#define WRAP0_CONTEXT_ARGUMENT1 YK0A_CONTEXT_GPR5
#define WRAP0_CONTEXT_ARGUMENT2 YK0A_CONTEXT_GPR6

#define WRAP0_PROCEDURE_ACCEPT 0
#define WRAP0_PROCEDURE_CALL 1
#define WRAP0_PROCEDURE_LISTEN 2

__inline__ static yk0a_error_t wrap0_accept(yk0a_port_t port, wrap_port_t wrap_port, yk0a_task_t *task) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t gpr3 __asm__("$3") = port;
	register yk0a_uint32_t gpr4 __asm__("$4") = WRAP0_PROCEDURE_ACCEPT;
	register yk0a_uint32_t gpr5 __asm__("$5") = wrap_port;
	__asm__ __volatile__("syscall" : "=r"(gpr2), "=r"(gpr3) : "0"(gpr2), "1"(gpr3), "r"(gpr4), "r"(gpr5) : "memory");
	if (task != NULL) {
		*task = gpr3;
	}
	return gpr2;
}

__inline__ static yk0a_error_t wrap0_call(yk0a_port_t port, wrap_port_t wrap_port) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t gpr3 __asm__("$3") = port;
	register yk0a_uint32_t gpr4 __asm__("$4") = WRAP0_PROCEDURE_CALL;
	register yk0a_uint32_t gpr5 __asm__("$5") = wrap_port;
	__asm__ __volatile__("syscall" : "=r"(gpr2) : "0"(gpr2), "r"(gpr3), "r"(gpr4), "r"(gpr5) : "memory");
	return gpr2;
}

__inline__ static yk0a_error_t wrap0_listen(yk0a_port_t port, wrap_port_t wrap_port, yk0a_flag_t flag_wait, yk0a_flag_t *flag_out) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t gpr3 __asm__("$3") = port;
	register yk0a_uint32_t gpr4 __asm__("$4") = WRAP0_PROCEDURE_LISTEN;
	register yk0a_uint32_t gpr5 __asm__("$5") = wrap_port;
	register yk0a_uint32_t gpr6 __asm__("$6") = flag_wait;
	__asm__ __volatile__("syscall" : "=r"(gpr2), "=r"(gpr3) : "0"(gpr2), "1"(gpr3), "r"(gpr4), "r"(gpr5), "r"(gpr6) : "memory");
	if (flag_out != NULL) {
		*flag_out = gpr3;
	}
	return gpr2;
}

#define WRAP1_CONTEXT_ARGUMENT0 YK0A_CONTEXT_GPR6
#define WRAP1_CONTEXT_ARGUMENT1 YK0A_CONTEXT_GPR7

#define WRAP1_PROCEDURE_CONS_GETCHAR 0
#define WRAP1_PROCEDURE_CONS_HALT 1
#define WRAP1_PROCEDURE_CONS_PUTCHAR 2

__inline__ static yk0a_error_t wrap1_cons_getchar(yk0a_port_t port, wrap_port_t wrap_port, y_uint8_t *c) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t gpr3 __asm__("$3") = port;
	register yk0a_uint32_t gpr4 __asm__("$4") = WRAP0_PROCEDURE_CALL;
	register yk0a_uint32_t gpr5 __asm__("$5") = wrap_port;
	register yk0a_uint32_t gpr6 __asm__("$6") = WRAP1_PROCEDURE_CONS_GETCHAR;
	__asm__ __volatile__("syscall" : "=r"(gpr2), "=r"(gpr3) : "0"(gpr2), "1"(gpr3), "r"(gpr4), "r"(gpr5), "r"(gpr6) : "memory");
	if (c != NULL) {
		*c = gpr3;
	}
	return gpr2;
}

__inline__ static yk0a_error_t wrap1_cons_halt(yk0a_port_t port, wrap_port_t wrap_port) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t gpr3 __asm__("$3") = port;
	register yk0a_uint32_t gpr4 __asm__("$4") = WRAP0_PROCEDURE_CALL;
	register yk0a_uint32_t gpr5 __asm__("$5") = wrap_port;
	register yk0a_uint32_t gpr6 __asm__("$6") = WRAP1_PROCEDURE_CONS_HALT;
	__asm__ __volatile__("syscall" : "=r"(gpr2) : "0"(gpr2), "r"(gpr3), "r"(gpr4), "r"(gpr5), "r"(gpr6) : "memory");
	return gpr2;
}

__inline__ static yk0a_error_t wrap1_cons_putchar(yk0a_port_t port, wrap_port_t wrap_port, y_uint8_t c) {
	register yk0a_uint32_t gpr2 __asm__("$2") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t gpr3 __asm__("$3") = port;
	register yk0a_uint32_t gpr4 __asm__("$4") = WRAP0_PROCEDURE_CALL;
	register yk0a_uint32_t gpr5 __asm__("$5") = wrap_port;
	register yk0a_uint32_t gpr6 __asm__("$6") = WRAP1_PROCEDURE_CONS_PUTCHAR;
	register yk0a_uint32_t gpr7 __asm__("$7") = c;
	__asm__ __volatile__("syscall" : "=r"(gpr2) : "0"(gpr2), "r"(gpr3), "r"(gpr4), "r"(gpr5), "r"(gpr6), "r"(gpr7) : "memory");
	return gpr2;
}

wrap_port_control_t wrap_port_control[WRAP_PORT_SIZE];
wrap_task_control_t wrap_task_control[YK0A_SETUP_TASK_SIZE];

extern y_uint8_t _gp[];

y_uint8_t task0_stack[TASK0_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));
y_uint8_t task1_stack[TASK1_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));
y_uint8_t task2_stack[TASK2_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));
y_uint8_t task3_stack[TASK3_STACK_SIZE] __attribute__((aligned(Y_STACK_ALIGNMENT)));

void task1(void);
void task2(void);
void task3(void);

void task0(void) {
	yk0a_flag_t f;
	yk0a_task_t t;
	y_device_cons_initialize();
	y_yk0a_start(1, task1, task1_stack + TASK1_STACK_SIZE / Y_STACK_ALIGNMENT * Y_STACK_ALIGNMENT - Y_STACK_ARGUMENT, _gp, y_mips_read_status());
	y_yk0a_start(2, task2, task2_stack + TASK2_STACK_SIZE / Y_STACK_ALIGNMENT * Y_STACK_ALIGNMENT - Y_STACK_ARGUMENT, _gp, y_mips_read_status());
	yk0a_set_priority(3, 1, NULL);
	y_yk0a_start(3, task3, task3_stack + TASK3_STACK_SIZE / Y_STACK_ALIGNMENT * Y_STACK_ALIGNMENT - Y_STACK_ARGUMENT, _gp, y_mips_read_status());
	yk0a_listen(YK0A_SETUP_EXCEPTION_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
	y_device_cons_put_string("{");
	y_device_cons_put_hexadecimal(f);
	if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_EXCEPTION_PORT, &t) == 0) {
		yk0a_context_value_t cause;
		yk0a_get_context(t, YK0A_CONTEXT_CAUSE, &cause);
		y_device_cons_put_string(" ");
		y_device_cons_put_hexadecimal(t);
		y_device_cons_put_string(" ");
		y_device_cons_put_hexadecimal(cause);
	}
	y_device_cons_put_string("}\n");
	y_device_cons_halt();
}

void task1(void) {
	wrap_port_t p;
	for (p = 0; p < WRAP_PORT_SIZE; ++p) {
		y_queue_initialize((y_queue_t *)&wrap_port_control[p]);
		wrap_port_control[p].flag = 0;
	}
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		yk0a_listen(YK0A_SETUP_INTERRUPT_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_INTERRUPT_PORT, &t) == 0) {
			yk0a_context_value_t a0;
			yk0a_context_value_t a1;
			yk0a_get_context(t, WRAP0_CONTEXT_ARGUMENT0, &a0);
			switch (a0) {
			case WRAP0_PROCEDURE_ACCEPT:
				yk0a_get_context(t, WRAP0_CONTEXT_ARGUMENT1, &a1);
				if (a1 >= WRAP_PORT_SIZE || (wrap_port_control[a1].flag & YK0A_FLAG_KERNEL_SENDER) == 0) {
					yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT0, 1, NULL);
					yk0a_reply(t);
				} else {
					yk0a_task_t replyee;
					replyee = (wrap_task_control_t *)y_queue_remove((y_queue_t *)&wrap_port_control[a1]) - &wrap_task_control[0];
					if (y_queue_copy((y_queue_t *)&wrap_port_control[a1]) == NULL) {
						wrap_port_control[a1].flag &= ~YK0A_FLAG_KERNEL_SENDER;
					}
					yk0a_connect(replyee, t);
					yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT1, replyee, NULL);
					yk0a_reply(t);
				}
				break;
			case WRAP0_PROCEDURE_CALL:
				yk0a_get_context(t, WRAP0_CONTEXT_ARGUMENT1, &a1);
				if (a1 >= WRAP_PORT_SIZE) {
					yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT0, 1, NULL);
					yk0a_reply(t);
				} else {
					y_queue_append((y_queue_t *)&wrap_port_control[a1], (y_queue_t *)&wrap_task_control[t]);
					wrap_port_control[a1].flag |= YK0A_FLAG_KERNEL_SENDER;
					if ((wrap_port_control[a1].flag & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
						yk0a_context_value_t a2;
						yk0a_get_context(wrap_port_control[a1].receiver, WRAP0_CONTEXT_ARGUMENT2, &a2);
						if ((a2 & YK0A_FLAG_KERNEL_SENDER) != 0) {
							yk0a_set_context(wrap_port_control[a1].receiver, YK0A_CONTEXT_ARGUMENT1, wrap_port_control[a1].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK), NULL);
							yk0a_reply(wrap_port_control[a1].receiver);
							wrap_port_control[a1].flag &= ~YK0A_FLAG_KERNEL_RECEIVER;
						}
					}
				}
				break;
			case WRAP0_PROCEDURE_LISTEN:
				yk0a_get_context(t, WRAP0_CONTEXT_ARGUMENT1, &a1);
				if (a1 >= WRAP_PORT_SIZE || (wrap_port_control[a1].flag & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
					yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT0, 1, NULL);
					yk0a_reply(t);
				} else {
					yk0a_context_value_t a2;
					yk0a_get_context(t, WRAP0_CONTEXT_ARGUMENT2, &a2);
					if ((a2 & (wrap_port_control[a1].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK) | YK0A_FLAG_KERNEL_RECEIVER)) != 0) {
						yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT1, wrap_port_control[a1].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK) | YK0A_FLAG_KERNEL_RECEIVER, NULL);
						yk0a_reply(t);
						wrap_port_control[0].flag &= ~a2 | YK0A_FLAG_KERNEL_MASK;
					} else {
						wrap_port_control[a1].flag |= YK0A_FLAG_KERNEL_RECEIVER;
						wrap_port_control[a1].receiver = t;
					}
				}
				break;
			default:
				yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT0, 1, NULL);
				yk0a_reply(t);
				break;
			}
		}
		if ((f & ~YK0A_FLAG_KERNEL_MASK) != 0) {
			wrap_port_control[0].flag |= f & ~YK0A_FLAG_KERNEL_MASK;
			for (p = 0; p < WRAP_PORT_SIZE; ++p) {
				if ((wrap_port_control[p].flag & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
					yk0a_context_value_t a2;
					yk0a_get_context(wrap_port_control[p].receiver, WRAP0_CONTEXT_ARGUMENT2, &a2);
					if ((a2 & wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK) != 0) {
						yk0a_set_context(wrap_port_control[p].receiver, YK0A_CONTEXT_ARGUMENT1, wrap_port_control[p].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK), NULL);
						yk0a_reply(wrap_port_control[p].receiver);
						wrap_port_control[0].flag &= ~a2 | YK0A_FLAG_KERNEL_MASK;
						wrap_port_control[p].flag &= ~YK0A_FLAG_KERNEL_RECEIVER;
					}
				}
			}
		}
	}
}

void task2(void) {
	y_uint8_t c;
	y_queue_t q;
	c = 0;
	y_queue_initialize(&q);
	yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		wrap0_listen(YK0A_SETUP_INTERRUPT_PORT, 0, YK0A_FLAG_KERNEL_SENDER | DEVICE_INTERRUPT_CONS, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && wrap0_accept(YK0A_SETUP_INTERRUPT_PORT, 0, &t) == 0) {
			yk0a_context_value_t a0;
			yk0a_context_value_t a1;
			yk0a_get_context(t, WRAP1_CONTEXT_ARGUMENT0, &a0);
			switch (a0) {
			case WRAP1_PROCEDURE_CONS_GETCHAR:
				y_queue_append(&q, (y_queue_t *)&wrap_task_control[t]);
				break;
			case WRAP1_PROCEDURE_CONS_PUTCHAR:
				yk0a_get_context(t, WRAP1_CONTEXT_ARGUMENT1, &a1);
				yk0a_reply(t);
				y_device_cons_putchar(a1);
				break;
			default:
				yk0a_forward(YK0A_SETUP_EXCEPTION_PORT, t);
				break;
			}
		}
		if ((f & DEVICE_INTERRUPT_CONS) != 0) {
			c = y_device_cons_getchar();
			yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
		}
		if (c != 0 && y_queue_copy(&q) != NULL) {
			t = (wrap_task_control_t *)y_queue_remove(&q) - &wrap_task_control[0];
			yk0a_set_context(t, YK0A_CONTEXT_ARGUMENT1, c, NULL);
			yk0a_reply(t);
			c = 0;
		}
	}
}

void task3(void) {
	while (1) {
		y_uint8_t c;
		wrap1_cons_getchar(YK0A_SETUP_INTERRUPT_PORT, 0, &c);
		if (c < ' ') {
			break;
		}
		wrap1_cons_putchar(YK0A_SETUP_INTERRUPT_PORT, 0, c);
	}
	wrap1_cons_halt(YK0A_SETUP_INTERRUPT_PORT, 0);
}
